-----------------------------------------------------------------------------------------------
-- Client Lua Script for InterruptThat
-- Copyright (c) legowxelab2z8. All rights reserved
-- Slightly edited by Cargo Cult for API10 compatibility
-----------------------------------------------------------------------------------------------

-----------------------------------------------------------------------------------------------
-- A word about variable naming
-- Abilities and Spells in WildStar can get messy and confusing.
-- In an effort to reduce confusion for myself and anyone who is reading this I decided to define some names.
--
-- Abilities - Abilities are different than Spells.  Abilities are what make up your limited action set
--	ActionSetLib.GetCurrentActionSet() Can be used to get all of the character's active Abilitiy Ids.
--	AbilityBook.GetAbilitiesList() can be used to lookup info for that ability.
--		I walk the list looking for Ids returned by GetCurrentActionSet looking for a match.
--	AbilityBook.GetAbilityInfo() seems to always return null, and I can't find a carbine example.
--
-- Spells - Spells are different than Abilities.
--	Spells can be retrieved by tAbility.tTiers[tAbility.nCurrentTier].splObject
--		tAbility comes from GetAbilitiesList()
--	There are multiple Spells for each Ability.  Most Abiltilities have 9 Spells ("Base" rank and Tier 1-8)
--		Base Tier is [1] and Tier 1 is [2] in the tAbility.tTiers table.
--	GameLib.GetSpell() Can be used to retrieve a splObject for a nSpellId
--
-- SpellEffect - What I call a SpellEffect is the spell that is exposed by tEventArgs.splCallingSpell
--	For example the Esper spell Crush Tier 1 has a SpellId of 52050 but the ability that triggers
--	OnCombatLogCCState is 52059.  Both spells have the same GetName() and GetRank() values but
--	GetBaseSpellId() returns 26488 which is the spell CreatureGib.
--	Using a SpellEffect when you intend to use a Spell will cause issues with retrieving some spell
--	information such as GetCooldownTime() that returns 0 for the SpellEffect.
--
--	Not all Spells have a SpellEffect.  Some Spells use the same Spell for tEventArgs.splCallingSpell
--
-- A spell with a SpellEffect will have 18 different Spell Ids and 1 Ability Id!
-----------------------------------------------------------------------------------------------


require "Window"
require "ICComm"
-----------------------------------------------------------------------------------------------
-- InterruptThat Module Definition
-----------------------------------------------------------------------------------------------
local InterruptThat = {}

-----------------------------------------------------------------------------------------------
-- Constants
local sVersion = "0.0.7"

local eCommMessageType =
  {
    LASChange = 1,
    CooldownChanged = 2,
    InformationRequest = 3,
  }

--This is a list of abilties that cause interrupts that we are interested in tracking
--Uses Ability.nId as key. (AbilityId not SpellId)
--Interrupts means that the ability will cause a CC effect on a target with 0 Interrupt Armor
--IARemovals is a table of Interrupt armor removed for each spell Tier.
--IARemovals[1] is Base Tier, IARemovals[2] is Tier1
local tInterruptSpellTable =
  {
    [19022] = {Interrupts = true, IARemovals = {1,1,1,1,2,2,2,2,2}}, 	-- Esper:Crush
    [19355] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,1}}, 	-- Esper:Incapacitate
    [19029] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,1}}, 	-- Esper:Shockwave
    [25635] = {Interrupts = true, IARemovals = {1,1,1,1,2,2,2,2,2}}, 	-- Engineer:Zap
    [34176] = {Interrupts = false, IARemovals = {1,1,1,1,1,1,1,1,1}}, 	-- Engineer:Obstruct Vision
    [34009] = {Interrupts = true, IARemovals = {0,0,0,0,0,0,0,0,0}}, 	-- Engineer:Blitz
    [20325] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,2}}, 	-- Spellslinger:Gate
    [30160] = {Interrupts = false, IARemovals = {1,1,1,1,2,2,2,2,2}}, 	-- Spellslinger:Arcane Shock
    [16454] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,1}}, 	-- Spellslinger:Spatial Shift
    [26543] = {Interrupts = true, IARemovals = {1,1,1,1,2,2,2,2,2}}, 	-- Medic:Paralytic Surge
    [26529] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,1}},  -- Medic:Magnetic Lockdown
    [23173] = {Interrupts = true, IARemovals = {1,1,1,1,2,2,2,2,2}}, 	-- Stalker:Stagger
    [23587] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,1}}, 	-- Stalker:False Retreat
    [23705] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,1}}, 	-- Stalker:Collapse
    [38017] = {Interrupts = true, IARemovals = {1,1,1,1,2,2,2,2,2}}, 	-- Warrior:Kick
    [18547] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,1}}, 	-- Warrior:Flash Bang
    [18363] = {Interrupts = true, IARemovals = {1,1,1,1,2,2,2,2,2}}, 	-- Warrior:Grapple
  }

--Used for Building SpellEffect Info for BuildSpellTable()
local tSpellNames =
  {
    ["Crush"] = true,
    ["Incapacitate"] = true,
    ["Shockwave"] = true,
    ["Zap"] = true,
    ["Obstruct Vision"] = true,
    ["Gate"] = true,
    ["Arcane Shock"] = true,
    ["Spatial Shift"] = true,
    ["Paralytic Surge"] = true,
    ["Magnetic Lockdown"] = true,
    ["Stagger"] = true,
    ["False Retreat"] = true,
    ["Collapse"] = true,
    ["Kick"] = true,
    ["Flash Bang"] = true,
    ["Grapple"] = true,
    ["Blitz"] = true,
  }

-----------------------------------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------------------------------

local Debug = false --Debugging flag for DebugPrint()

local bSendInfoRequest = false

--Used to store the last time the interrupt was cast
--Also used so we dont show a message more than once for aoe interrupts
--[Player][nSpellId] = {timeLastCast, fCooldown}
local tLastCast = {}

local tSavedData --Used to store save data. Needed if OnDocLoaded hasn't happened yet.
local bDocLoaded = false -- Used to track if OnDocLoaded has been called.
local bRestored = false -- Used to track if OnRestore has been called.

local tSpellConversions = {} --Used for storing SpellEffect Info for BuildSpellTable()

local LastUpdateSent = nil  --Used for tracking the last time we sent a update comm message so requests dont cause us to spam.

local LastLeader = nil


function InterruptThat:OnSave(eType)
  if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
  local tSaved = {
    tGroupFrameData = {
      AnchorOffsets = {self.wndGroupFrame:GetAnchorOffsets()}
    }
  }
  return tSaved
end

-----------------------------------------------------------------------------------------------
-- OnRestore is not called the first time the addon is loaded.
-- Can be called before or after OnDocLoaded.
-----------------------------------------------------------------------------------------------
function InterruptThat:OnRestore(eType, tRestoreData)
  tSavedData = tRestoreData
  bRestored = true
  if bDocLoaded then --Frames are created.
    self:RestoreGroupFrame()
  end
end

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function InterruptThat:new(o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self

  -- initialize variables here
  self.tCurrentAbilities = {}
  self.tInterruptAbilities = {} --Used for tracking your own interrupt abilities.
  self.tCooldownTimers = {}
  self.PlayerFrames = {}
  self.CommChannel = nil
  self.CommPlayers = {}
  return o
end

function InterruptThat:Init()
  LibJSON = Apollo.GetPackage("Lib:dkJSON-2.5").tPackage

  local bHasConfigureFunction = false
  local strConfigureButtonText = ""
  local tDependencies = {
    -- "UnitOrPackageName",
    }
  Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end


-----------------------------------------------------------------------------------------------
-- InterruptThat OnLoad
-----------------------------------------------------------------------------------------------
function InterruptThat:OnLoad()
  -- load our form file
  self.xmlDoc = XmlDoc.CreateFromFile("InterruptThat.xml")
  self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end



-----------------------------------------------------------------------------------------------
-- InterruptThat OnDocLoaded
-----------------------------------------------------------------------------------------------
function InterruptThat:OnDocLoaded()

  if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
    self.wndMain = Apollo.LoadForm(self.xmlDoc, "InterruptThatForm", nil, self)
    if self.wndMain == nil then
      Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
      return
    end

    bDocLoaded = true

    self.wndMain:Show(false, true)

    self.wndGroupFrame = Apollo.LoadForm(self.xmlDoc, "GroupFrame", nil, self)
    self.wndGroupFrame:Show(true, true)

    if bRestored then --Saved data has already been loaded
      self:RestoreGroupFrame()
    end

    --add the player  (Bugs when first loading since addon loads before GetPlayerUnit will work)
    --self:AddPlayerFrame(GameLib.GetPlayerUnit():GetName())

    -- if the xmlDoc is no longer needed, you should set it to nil
    -- self.xmlDoc = nil

    -- Register handlers for events, slash commands and timer, etc.
    Apollo.RegisterEventHandler("CombatLogCCState", "OnCombatLogCCState", self)
    Apollo.RegisterSlashCommand("it", "OnInterruptThatOn", self)
    Apollo.RegisterEventHandler("AbilityBookChange", "LASChanged", self)

    Apollo.RegisterEventHandler("Group_Updated", "OnGroupUpdated", self)
    Apollo.RegisterEventHandler("Group_Remove",	"OnGroupRemove", self)				-- ( name, reason )
    Apollo.RegisterEventHandler("Group_Add", "OnGroupAdd", self)					-- ( name )
    Apollo.RegisterEventHandler("Group_Left", "OnGroupLeft", self)				-- ( reason )


    Apollo.RegisterEventHandler("NextFrame","NextFrame", self) --Poll for changes to cooldowns, there is no reliable event :(


    -- Do additional Addon initialization here
  end
end

-----------------------------------------------------------------------------------------------
-- InterruptThat Functions
-----------------------------------------------------------------------------------------------
function InterruptThat:OnCombatLogCCState(tEventArgs)
  if not tEventArgs then return end
  if not tEventArgs.unitTarget then return end
  if not tEventArgs.unitCaster then return end
  if tEventArgs.bRemoved then return end -- indicates that the cc effect ended.


  Event_FireGenericEvent("SendVarToRover", "LogCCState tEventArgs", tEventArgs, 0)

  local unitCaster = (tEventArgs.unitCasterOwner or tEventArgs.unitCaster)
  local strCaster = unitCaster:GetName()
  local strTarget = tEventArgs.unitTarget:GetName()
  local unitPlayer = GameLib.GetPlayerUnit()
  local splSpell = tEventArgs.splCallingSpell
  local nSpellId = splSpell:GetId()
  if not (unitCaster == GameLib.GetPlayerUnit() or unitCaster:IsInYourGroup()) then return end
  if self.CommPlayers[strCaster] then return end --This player has the addon, let them report instead of attempting to track

  --self:DebugPrint(strCaster .. " " .. splSpell:GetName() .. " " .. nSpellId .. " " .. splSpell:GetBaseSpellId())

  --The spell that shows up isn't always the same as the spell that is cast, so convert to spell cast if needed.
  if self.tSpellEffectIdConversion[nSpellId] then
    --self:DebugPrint("Converting ID " .. nSpellId .. " to " .. self.tSpellEffectIdConversion[nSpellId])
    nSpellId = self.tSpellEffectIdConversion[nSpellId]
    splSpell = GameLib.GetSpell(nSpellId)
  end

  local stunned = tEventArgs.unitTarget:GetCCStateTimeRemaining(tEventArgs.eState) > 0
  if stunned then
  --self:DebugPrint(strTarget .. " stunned.")
  end


  if tInterruptSpellTable[splSpell:GetBaseSpellId()] then
    --self:DebugPrint("Tracking " .. splSpell:GetName())
    self:TrackSpell(unitCaster:GetName(), splSpell)
  else
  --[[

		tInterruptSpellTable[splSpell:GetBaseSpellId()] = {Interrupts = true, IARemovals = {1,1,1,1,1,1,1,1,1}}

		tInterruptSpellTable[splSpell:GetBaseSpellId()].IARemovals[splSpell:GetTier()] = tEventArgs.nInterruptArmorHit

		self:TrackSpell(unitCaster:GetName(), splSpell)

		self:DebugPrint(strCaster 

			.. " " .. splSpell:GetName() 

			.. " Id " .. nSpellId

			.. " Base Id " .. splSpell:GetBaseSpellId())

		]]
  end
end

--When AbilityBookChange event fires AbilityBook.GetAbilitiesList() doesn't always return the new list.
--Set a timer to give the game a chance to update the LAS before we grab it.
function InterruptThat:LASChanged()
  --self:DebugPrint("LASChanged")
  self.RebuildLASTimer = ApolloTimer.Create(.1, false, "RebuildLAS", self)
end

function InterruptThat:RebuildLAS()
  if GameLib.GetPlayerUnit() == nil then self:LASChanged() return end --Set the timer again if not loaded yet.
  self:DebugPrint("RebuildLAS")
  local unitPlayer = GameLib.GetPlayerUnit()
  local tActionSetAbilties = ActionSetLib.GetCurrentActionSet()
  if not tActionSetAbilties then return end
  local tAllAbilities = AbilityBook.GetAbilitiesList()
  local tAbiltyIndexes = {}
  self.tInterruptAbilities = {}

  for indx = 1, 8 do
    if tActionSetAbilties[indx] > 0 then
      tAbiltyIndexes[tActionSetAbilties[indx]] = indx
    end
  end
  --self.wndMain:FindChild("EditBox"):SetText("")
  local strLASText = ""
  self.tCurrentAbilities = {}
  for i, tAbility in pairs(tAllAbilities) do
    if tAbility.nId and tAbiltyIndexes[tAbility.nId] then
      self.tCurrentAbilities[tAbiltyIndexes[tAbility.nId]] = tAbility
      strLASText = strLASText .. tAbility.strName .. " " .. tAbility.tTiers[tAbility.nCurrentTier].splObject:GetId() .. "\n"
    end
  end
  --self.wndMain:FindChild("EditBox"):SetText(strLASText)

  for i, tAbility in pairs(self.tCurrentAbilities) do
    --self:DebugPrint(tAbility.strName .. " " .. tAbility.nId)
    if tInterruptSpellTable[tAbility.nId] then
      local nCooldown = tAbility.tTiers[tAbility.nCurrentTier].splObject:GetCooldownTime()
      local strCooldown = string.format("%.2f", nCooldown)
      local nSpellId = tAbility.tTiers[tAbility.nCurrentTier].splObject:GetId()
      self:DebugPrint("Spell: " .. tAbility.strName
        .. " Tier: " .. tAbility.nCurrentTier
        .. " Cooldown: " .. strCooldown
        .. " Removes: " .. tInterruptSpellTable[tAbility.nId].IARemovals[tAbility.nCurrentTier])

      self.tInterruptAbilities[tAbility.nId] = tAbility
      --self.tCooldownTimers[tAbility] = 0
      --self:AddPlayerSpell(unitPlayer:GetName(), nSpellId)
    end
  end
  self:SendCurrentInterrupts()
end

function InterruptThat:NextFrame()

  if bSendInfoRequest then
    local bSuccess = self.CommChannel:SendMessage(LibJSON.encode({eType = eCommMessageType.InformationRequest}))
    if bSuccess then
      bSendInfoRequest = false
      self:RebuildLAS()
    end
  end

  --Get your own cooldowns
  for iAbilityId, tAbility in pairs(self.tInterruptAbilities) do
    local splCurrentSpell = tAbility.tTiers[tAbility.nCurrentTier].splObject
    local nSpellId = splCurrentSpell:GetId()
    local fCooldown = splCurrentSpell:GetCooldownRemaining()
    local fOldCooldown = self.tCooldownTimers[tAbility] or 0
    if fOldCooldown < fCooldown then --Cooldown increased
      self:DebugPrint("Spell: " .. tAbility.strName
        .. " used. Cooldown: " .. fCooldown .. ".")
    self:SendCooldownInformation(nSpellId, fCooldown)
    elseif fOldCooldown - fCooldown > 1 then -- Cooldown changed a lot, report it again
      self:DebugPrint("Spell: " .. tAbility.strName
        .. " Cooldown changed. New Cooldown: " .. fCooldown .. ".")
    self:SendCooldownInformation(nSpellId, fCooldown)
    end
    self.tCooldownTimers[tAbility] = fCooldown
  end

  --Draw cooldown timers
  for strPlayerName,tSpells in pairs(tLastCast) do
    if self.PlayerFrames[strPlayerName] then
      local wndPlayerFrame = self.PlayerFrames[strPlayerName]
      for nSpellId,CooldownInfo in pairs(tSpells) do
        --self:DebugPrint(nSpellId)
        local wndSpellFrame = wndPlayerFrame:FindChild(nSpellId)
        if wndSpellFrame then
          --for k,v in pairs(CooldownInfo) do self:DebugPrint(k .. " " .. v) end
          local nCooldown = CooldownInfo.nCooldown - Time.SecondsElapsed(CooldownInfo.timer)
          if nCooldown > 0 then
            local strCooldown
            if nCooldown > 10 then
              strCooldown = string.format("%d", nCooldown)
            else
              strCooldown = string.format("%.1f", nCooldown)
            end
            wndSpellFrame:SetText(strCooldown)
          else
            wndSpellFrame:SetText("")
            tLastCast[strPlayerName][nSpellId] = nil
            wndSpellFrame:FindChild("Overlay"):SetOpacity(0,1)
          end
        end
      end
    end
  end
end

function InterruptThat:TrackSpell(strCaster, splSpell)
  local nCooldown = splSpell:GetCooldownTime()
  local nSpellId = splSpell:GetId()
  if not tLastCast[strCaster] then tLastCast[strCaster] = {} end
  if not tLastCast[strCaster][nSpellId] then tLastCast[strCaster][nSpellId] = {} end
  tLastCast[strCaster][nSpellId] = {timer = Time.Now(),nCooldown = nCooldown}

  self:AddPlayerSpell(strCaster, nSpellId)
  local wndPlayerFrame = self.PlayerFrames[strCaster]
  if not wndPlayerFrame then self:DebugPrint("No Player frame" .. strCaster) return end
  local wndSpellFrame = wndPlayerFrame:FindChild(nSpellId)
  if not wndSpellFrame then self:DebugPrint("No spell frame" .. strCaster .. " " .. nSpellId) return end
  wndSpellFrame:FindChild("Overlay"):SetOpacity(.7,0)
end

function InterruptThat:TrackReportedSpell(sPlayerName, nSpellId, fCooldown)
  if not tLastCast[sPlayerName] then tLastCast[sPlayerName] = {} end
  if not tLastCast[sPlayerName][nSpellId] then tLastCast[sPlayerName][nSpellId] = {} end
  tLastCast[sPlayerName][nSpellId] = {timer = Time.Now(),nCooldown = fCooldown}
  self:AddPlayerSpell(sPlayerName, nSpellId)
  local wndPlayerFrame = self.PlayerFrames[sPlayerName]
  if not wndPlayerFrame then self:DebugPrint("No Player frame" .. sPlayerName) return end
  local wndSpellFrame = wndPlayerFrame:FindChild(nSpellId)
  if not wndSpellFrame then self:DebugPrint("No spell frame" .. sPlayerName .. " " .. nSpellId) return end
  wndSpellFrame:FindChild("Overlay"):SetOpacity(.7,0)
end


function InterruptThat:OnGroupUpdated()
  for idx = 1, GroupLib.GetMemberCount() do
    local tMemberInfo = GroupLib.GetGroupMember(idx)
    if tMemberInfo ~= nil then
      self:AddPlayerFrame(tMemberInfo.strCharacterName)
    end
  end
  self:GetGroupLeader()
end

function InterruptThat:OnGroupAdd(sName)
  if self:AddPlayerFrame(sName) then
    self:DebugPrint("OnGroupAdd - " .. sName .. " added.")
  end
end

--Fires when a player leaves the group
function InterruptThat:OnGroupRemove(sName, eReason)
  self:DebugPrint("OnGroupRemove")
  self:RemovePlayerFrame(sName)
end

--Fires when the player leaves a group
function InterruptThat:OnGroupLeft(eReason)
  self:DebugPrint("OnGroupLeft")
  local sMyName = GameLib.GetPlayerUnit():GetName()
  for sPlayerName, wndPlayerFrame in pairs(self.PlayerFrames) do
    self:RemovePlayerFrame(sPlayerName)
  end
  self.CommChannel = nil
  LastLeader = nil
end

function InterruptThat:GetGroupLeader()
  if not GroupLib.InGroup() then LastLeader = nil return end
  local sLeader = nil
  for idx = 1, GroupLib.GetMemberCount() do
    local tMemberInfo = GroupLib.GetGroupMember(idx)
    if tMemberInfo ~= nil then
      if tMemberInfo.bIsLeader then sLeader = tMemberInfo.strCharacterName end
    end
  end
  if sLeader ~= LastLeader then -- Leader changed
    self:GroupLeaderChanged(sLeader)
  end
  LastLeader = sLeader
  return sLeader
end

function InterruptThat:GroupLeaderChanged(sLeaderName)
  self:DebugPrint("GroupLeaderChanged")
  self:DebugPrint("Joinning Channel " .. sLeaderName)
  self.CommChannel = ICCommLib.JoinChannel("InterruptThat" .. sLeaderName, ICCommLib.CodeEnumICCommChannelType.Group)
  self.CommChannel:SetReceivedMessageFunction("OnChannelMessage", self)
  bSendInfoRequest = true
end

function InterruptThat:OnChannelMessage(channel, strMsg, sSenderName)
  self:DebugPrint("OnChannelMessage " .. strMsg)
  local tMsg = LibJSON.decode(strMsg)

  if not tMsg then return end
  Event_FireGenericEvent("SendVarToRover", "ChannelMessage", tMsg, 0)
  if self.PlayerFrames[sSenderName] then --They are in our party, process message.
    --self:DebugPrint("Message from " .. sSenderName)
    if not self.CommPlayers[sSenderName] then --First time we got a message from them.
      self.CommPlayers[sSenderName] = true
      self.PlayerFrames[sSenderName]:FindChild("Name"):SetTextColor(ApolloColor.new("green"))
  end

  if tMsg.eType == eCommMessageType.LASChange then
    self:DebugPrint(sSenderName .. " LASChange Message")
    self:RemoveAllPlayerSpells(sSenderName)
    for idx, nSpellId in ipairs(tMsg) do
      self:AddPlayerSpell(sSenderName, nSpellId)
    end
  elseif tMsg.eType == eCommMessageType.CooldownChanged then
    self:DebugPrint("Cooldown Info Recieved " .. sSenderName)
    self:TrackReportedSpell(sSenderName, tMsg.tCooldownInfo.nSpellId, tMsg.tCooldownInfo.fCooldown)
  elseif tMsg.eType == eCommMessageType.InformationRequest then
    self:DebugPrint("InformationRequest recieved " .. sSenderName)
    self:RebuildLAS() --Triggers a send message
  end
  end
end
  
function InterruptThat:SendCurrentInterrupts()
  if not GroupLib.InGroup() or not self.CommChannel then return end
  self:DebugPrint("Sending Interrupts")
  local tInterruptSpells = {eType = eCommMessageType.LASChange}
  for iAbilityId, tAbility in pairs(self.tInterruptAbilities) do
    local splCurrentSpell = tAbility.tTiers[tAbility.nCurrentTier].splObject
    local nSpellId = splCurrentSpell:GetId()
    if nSpellId then
      table.insert(tInterruptSpells, nSpellId)
    end
  end
  local strMessage = LibJSON.encode(tInterruptSpells)

  self.CommChannel:SendMessage(strMessage)
  self:OnChannelMessage(self.CommChannel, strMessage, GameLib.GetPlayerUnit():GetName())
  Event_FireGenericEvent("SendVarToRover", "tInterruptSpells", tInterruptSpells, 0)
end

function InterruptThat:SendCooldownInformation(nSpellId, fCooldown)
  if not nSpellId then return end
  if not fCooldown then return end
  if not GroupLib.InGroup() or not self.CommChannel then return end
  self:DebugPrint("Sending cooldown " .. nSpellId)
  local tCooldownMessage = {
    eType = eCommMessageType.CooldownChanged,
    tCooldownInfo = {nSpellId = nSpellId, fCooldown = fCooldown},
  }
  
  local strMessage = LibJSON.encode(tCooldownMessage)
  
  self.CommChannel:SendMessage(strMessage)
  self:OnChannelMessage(self.CommChannel, strMessage, GameLib.GetPlayerUnit():GetName())
end

--Builds the table of spell casts and effects by walking though all the spell ids
--Only works for english localization
--Finds more Effect Ids than are needed, but I don't know a pragmatic way to filter them yet.
--Since we only care about player casts erroneous effect ids (such as a boss attack called Crush duplicating the Esper spell)
--Hopefully won't cause issues.
function InterruptThat:BuildSpellTable()
  for nSpellId = 1, 100000 do
    local splSpell = GameLib.GetSpell(nSpellId)
    if splSpell then
      local strName = splSpell:GetName()
      if tSpellNames[strName] then --Spell is an interrupt we want to track
        local nBaseSpell = splSpell:GetBaseSpellId()
        local nTier = splSpell:GetTier()

        if not tSpellConversions[strName] then tSpellConversions[strName] = {} end
        if not tSpellConversions[strName][nTier] then tSpellConversions[strName][nTier] = {} end

        if tInterruptSpellTable[nBaseSpell] then --Spell is the actual cast version of the spell not the effect
          tSpellConversions[strName][nTier].Cast = nSpellId
        else
          if not tSpellConversions[strName][nTier].Effect then tSpellConversions[strName][nTier].Effect = {} end
          table.insert(tSpellConversions[strName][nTier].Effect, nSpellId)
          --tSpellConversions[strName][nTier].Effect = nSpellId
        end
      end
    end
  end
  Event_FireGenericEvent("SendVarToRover", "tSpellConversions", tSpellConversions, 0)

  local strConversionTables = "local tSpellEffectIdConversion = \n{\n"
  for strSpellName,tConversionTable in pairs(tSpellConversions) do
    for nTier, tSpellInfo in pairs(tConversionTable) do
      local nCastId = tSpellInfo.Cast or 0
      if tSpellInfo.Effect and type(tSpellInfo.Effect) == "table" then
        for indx, nEffectId in pairs(tSpellInfo.Effect) do
          strConversionTables = strConversionTables .. "\t[" .. nEffectId .. "] = " .. nCastId .. ", --" .. strSpellName .. ":Tier" ..nTier-1 .."\n"
        end
      end
    end
  end
  strConversionTables = strConversionTables .. "}"
  self.wndMain:Invoke() -- show the window
  
  
  self.wndMain:FindChild("EditBox"):SetText(strConversionTables)

end

-- on SlashCommand "/it"
function InterruptThat:OnInterruptThatOn()

  self.wndMain:Invoke() -- show the window
   for nSpellId = 1, 100000 do
    local splSpell = GameLib.GetSpell(nSpellId)
    if splSpell then
      local strName = splSpell:GetName()
      if tSpellNames[strName] then --Spell is an interrupt we want to track
        local nBaseSpell = splSpell:GetBaseSpellId()
        local nTier = splSpell:GetTier()

        if not tSpellConversions[strName] then tSpellConversions[strName] = {} end
        if not tSpellConversions[strName][nTier] then tSpellConversions[strName][nTier] = {} end

        if tInterruptSpellTable[nBaseSpell] then --Spell is the actual cast version of the spell not the effect
          tSpellConversions[strName][nTier].Cast = nSpellId
        else
          if not tSpellConversions[strName][nTier].Effect then tSpellConversions[strName][nTier].Effect = {} end
          table.insert(tSpellConversions[strName][nTier].Effect, nSpellId)
          --tSpellConversions[strName][nTier].Effect = nSpellId
        end
      end
    end
  end
  Event_FireGenericEvent("SendVarToRover", "tSpellConversions", tSpellConversions, 0)

  local strConversionTables = "local tSpellEffectIdConversion = \n{\n"
  for strSpellName,tConversionTable in pairs(tSpellConversions) do
    for nTier, tSpellInfo in pairs(tConversionTable) do
      local nCastId = tSpellInfo.Cast or 0
      if tSpellInfo.Effect and type(tSpellInfo.Effect) == "table" then
        for indx, nEffectId in pairs(tSpellInfo.Effect) do
          strConversionTables = strConversionTables .. "\t[" .. nEffectId .. "] = " .. nCastId .. ", --" .. strSpellName .. ":Tier" ..nTier-1 .."\n"
        end
      end
    end
  end
  strConversionTables = strConversionTables .. "}"
  
   self.wndMain:FindChild("EditBox"):SetText(strConversionTables)
  
end

function InterruptThat:DebugPrint(strDebugString)
  if Debug then
    Print(strDebugString)
  end
  --ChatSystemLib.Command("/p " .. strDebugString)
end


-----------------------------------------------------------------------------------------------
-- InterruptThatForm Functions
-----------------------------------------------------------------------------------------------

function InterruptThat:AddPlayerFrame(strPlayerName)
  if not strPlayerName then return false end
  if self.PlayerFrames[strPlayerName] then return end

  local wndPlayerFrame = Apollo.LoadForm(self.xmlDoc, "PlayerForm", self.wndGroupFrame, self)

  wndPlayerFrame:SetName(strPlayerName)
  wndPlayerFrame:FindChild("Name"):SetText(string.sub(strPlayerName,1,7))

  self.PlayerFrames[strPlayerName] = wndPlayerFrame

  self.wndGroupFrame:ArrangeChildrenHorz()
  self:ResizeGroupFrame()
  return true
end

function InterruptThat:RemovePlayerFrame(strPlayerName)
  local wndPlayerFrame = self.PlayerFrames[strPlayerName]
  if not wndPlayerFrame then return end
  wndPlayerFrame:Destroy()
  self.PlayerFrames[strPlayerName] = nil
  self.CommPlayers[strPlayerName] = nil
  self:ResizeGroupFrame()
end

function InterruptThat:AddPlayerSpell(strPlayerName, nSpellId)
  --self:DebugPrint("AddPlayerSpell")
  if not nSpellId then return end
  local splSpell = GameLib.GetSpell(nSpellId)
  if not splSpell then return end
  local nBaseSpell = splSpell:GetBaseSpellId()
  if not self.PlayerFrames[strPlayerName] then self:AddPlayerFrame(strPlayerName) end
  local wndPlayerFrame = self.PlayerFrames[strPlayerName]

  --Make/Update the spellframe
  local wndSpellFrame = wndPlayerFrame:FindChild(nSpellId) --Try to get the frame if it already exists
  if wndSpellFrame == nil then
    wndSpellFrame = Apollo.LoadForm(self.xmlDoc, "CooldownIconForm", wndPlayerFrame, self)
  end
  wndSpellFrame:SetName(nSpellId)
  local strSpellIcon = splSpell:GetIcon()
  if not strSpellIcon then return end
  wndSpellFrame:SetSprite(strSpellIcon)
  wndSpellFrame:FindChild("Overlay"):SetOpacity(0,1)
  --self:DebugPrint(splSpell:GetBaseSpellId() .. " " .. splSpell:GetTier())
  local nTier = splSpell:GetTier()
  local sTier = ""
  if nTier == 1 then
    sTier = "Base"
  else
    sTier = "Tier: " .. tostring(nTier - 1)
  end

  wndSpellFrame:SetTooltip(splSpell:GetName() .. " " ..  sTier)

  --Add the Interrupt Armor Removal number
  local nIARemovals = "?"
  if tInterruptSpellTable[splSpell:GetBaseSpellId()] and
    tInterruptSpellTable[splSpell:GetBaseSpellId()].IARemovals[splSpell:GetTier()] then
    nIARemovals = tInterruptSpellTable[splSpell:GetBaseSpellId()].IARemovals[splSpell:GetTier()]
  end
  if not nIARemovals then nIARemovals = "?" end
  --self:DebugPrint("Removes " .. nIARemovals)
  wndSpellFrame:FindChild("IARemovalsForm"):SetText(nIARemovals)


  --Remove old spell ranks
  --Needed when tracking spells from other players when they don't have the addon to self report.
  local twndFrames = wndPlayerFrame:GetChildren()
  for indx, wndFrame in pairs(twndFrames) do
    local sName = wndFrame:GetName()
    local nFrameId = tonumber(sName)
    if nFrameId and nFrameId ~= nSpellId then
      local splFrameSpell = GameLib.GetSpell(nFrameId)
      if splFrameSpell then
        if splFrameSpell:GetBaseSpellId() == nBaseSpell then --Matchin base spells means duplicate (different tiers)
          self:RemovePlayerSpell(strPlayerName, nFrameId)
        end
      end
    end
  end

  --Update the height
  self:ResizePlayerFrame(wndPlayerFrame)

end

function InterruptThat:RemovePlayerSpell(strPlayerName, nSpellId)
  local wndPlayerFrame = self.PlayerFrames[strPlayerName]
  if not wndPlayerFrame then return end
  local wndSpellFrame = wndPlayerFrame:FindChild(nSpellId) --Try to get the frame if it already exists
  if not wndSpellFrame then return end
  wndSpellFrame:Destroy()
  self:ResizePlayerFrame(wndPlayerFrame)
end

function InterruptThat:RemoveAllPlayerSpells(strPlayerName)
  local wndPlayerFrame = self.PlayerFrames[strPlayerName]
  if not wndPlayerFrame then return end
  for idx, wndSpellFrame in pairs(wndPlayerFrame:GetChildren()) do
    local nSpellId = tonumber(wndSpellFrame:GetName()) or 0
    if GameLib.GetSpell(nSpellId) then
      self:RemovePlayerSpell(strPlayerName, nSpellId)
    end
  end
  self:ResizePlayerFrame(wndPlayerFrame)
end

function InterruptThat:ResizePlayerFrame(wndFrame)
  local left,top,right,bottom = wndFrame:GetAnchorOffsets()
  local height = 0
  for k,v in pairs(wndFrame:GetChildren()) do
    height = height + v:GetHeight()
  end
  wndFrame:SetAnchorOffsets(left,top,right,top+height)

  wndFrame:ArrangeChildrenVert(1)
  self:ResizeGroupFrame()
end

function InterruptThat:ResizeGroupFrame()
  local width = 0
  local height = 0
  for k,wndPlayer in pairs(self.wndGroupFrame:GetChildren()) do
    width = width + wndPlayer:GetWidth()
    height = math.max(height, wndPlayer:GetHeight())
  end
  local left,top,right,bottom = self.wndGroupFrame:GetAnchorOffsets()
  self.wndGroupFrame:SetAnchorOffsets(left,top,left+width,top+height)
end

function InterruptThat:RestoreGroupFrame()
  if bRestored then --Cant restore data if the data hasn't loaded
    local tGroupFrameData = tSavedData.tGroupFrameData
    if not tGroupFrameData or type(tGroupFrameData) ~= "table" then return end
    if tGroupFrameData.AnchorOffsets then
      self.wndGroupFrame:SetAnchorOffsets(unpack(tSavedData.tGroupFrameData.AnchorOffsets))
    end
  end
end

-- when the OK button is clicked
function InterruptThat:OnOK()
  self.wndMain:Close() -- hide the window
end

-- when the Cancel button is clicked
function InterruptThat:OnCancel()
  self.wndMain:Close() -- hide the window
end

function InterruptThat:CooldownIconFormOnMouseButtonUp(wndHandler, wndControl, eMouseButton)
  if eMouseButton == GameLib.CodeEnumInputMouse.Right then
    --clunky and should probably be updated
    local nSpellId = tonumber(wndControl:GetName())
    if nSpellId then
      local sPlayerName = wndControl:GetParent():FindChild("Name"):GetText()
      --self:DebugPrint("Mouse Up " .. wndControl:GetName() .. " " .. sPlayerName)
      if sPlayerName then
        self:RemovePlayerSpell(sPlayerName, nSpellId)
      end
    end
  end
end

function InterruptThat:BuildSpellDebugEditBox()

end

-----------------------------------------------------------------------------------------------
-- InterruptThat Instance
-----------------------------------------------------------------------------------------------
local InterruptThatInst = InterruptThat:new()
InterruptThatInst:Init()
